<div class="row">
    <div class="col-12">
        <div class="page-title-box">
            <div class="page-title-right">
                <ol class="breadcrumb m-0">
                    {% for crumb in breadcrumbs %}
                        <li class="breadcrumb-item">
                            <a href="{% url crumb.0 %}">{{ crumb.1 }}</a>
                        </li>
                    {% endfor %}
                    <li class="breadcrumb-item active">{{ breadcrumb_active }}</li>
                </ol>
            </div>
            <h4 class="page-title">
                {{ page_title }}
                {% if page_title_icon %}<i class="mdi {{ page_title_icon }} me-1"></i>{% endif %}
            </h4>
        </div>
    </div>
</div>
