# Generated by Django 4.2.7 on 2024-02-26 19:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("badelf", "0002_badelf_electrides_per_formula"),
    ]

    operations = [
        migrations.AddField(
            model_name="badelf",
            name="electrides_per_reduced_formula",
            field=models.FloatField(blank=True, null=True),
        ),
    ]
